@@ -14,7 +14,6 @@
 
 from jax import core
 from jax import numpy as jnp
-from jax import lazy
 from jax.interpreters import xla
 from jax.lib import xla_client
 from jax.lib import xla_bridge
@@ -41,9 +40,8 @@ def to_dlpack(x: xla.DeviceArrayProtocol, take_ownership: bool = False):
   if not isinstance(x, xla.DeviceArray):
     raise TypeError("Argument to to_dlpack must be a DeviceArray, got {}"
                     .format(type(x)))
-  buf = xla._force(x).device_buffer
   return xla_client._xla.buffer_to_dlpack_managed_tensor(
-      buf, take_ownership=take_ownership)
+      x.device_buffer, take_ownership=take_ownership)
 
 def from_dlpack(dlpack, backend=None):
   """Returns a `DeviceArray` representation of a DLPack tensor `dlpack`.
@@ -62,4 +60,4 @@ def from_dlpack(dlpack, backend=None):
   xla_shape = buf.xla_shape()
   assert not xla_shape.is_tuple()
   aval = core.ShapedArray(xla_shape.dimensions(), xla_shape.numpy_dtype())
-  return xla.make_device_array(aval, buf.device(), lazy.array(aval.shape), buf)  # pytype: disable=attribute-error
+  return xla.make_device_array(aval, buf.device(), buf)  # pytype: disable=attribute-error
