@@ -17,7 +17,7 @@ import numpy as np
 
 from jax._src.numpy import lax_numpy as jnp
 from jax._src.numpy.vectorize import vectorize
-from jax import ad_util
+from jax._src import ad_util
 from jax._src import api
 from jax import lax
 from jax import ops
@@ -99,13 +99,13 @@ def eigh(x, lower: bool = True, symmetrize_input: bool = True):
       eigendecomposition by computing :math:`\\frac{1}{2}(x + x^H)`.
 
   Returns:
-    A tuple ``(v, w)``.
-
-    ``v`` is an array with the same dtype as ``x`` (or its real counterpart if
-    complex) with shape ``[..., n]`` containing the eigenvalues of ``x``.
+    A tuple ``(w, v)``.
 
     ``w`` is an array with the same dtype as ``x`` such that ``w[..., :, i]`` is
     the eigenvector corresponding to ``v[..., i]``.
+
+    ``v`` is an array with the same dtype as ``x`` (or its real counterpart if
+    complex) with shape ``[..., n]`` containing the eigenvalues of ``x``.
   """
   if symmetrize_input:
     x = symmetrize(x)
@@ -475,11 +475,8 @@ def eigh_translation_rule(c, operand, lower):
   shape = c.get_shape(operand)
   dims = shape.dimensions()
   if dims[-1] == 0:
-    return xops.Tuple(c, [operand, xops.Reshape(operand, dims[:-1])])
-  if not lower:
-    n = len(dims)
-    operand = xops.Transpose(operand, list(range(n - 2)) + [n - 1, n - 2])
-  return xops.Tuple(c, xops.Eigh(operand))
+    return xops.Tuple(c, [operand, xops.Real(xops.Reshape(operand, dims[:-1]))])
+  return xops.Tuple(c, xops.Eigh(operand, lower=lower))
 
 def eigh_abstract_eval(operand, lower):
   if isinstance(operand, ShapedArray):
@@ -508,14 +505,6 @@ def _eigh_cpu_gpu_translation_rule(syevd_impl, c, operand, lower):
                            _nan_like(c, w))
   return xops.Tuple(c, [v, w])
 
-def _eigh_tpu_translation_rule(c, operand, lower):
-  # Fail gracefully for complex dtype (unsupported on TPU).
-  shape = c.get_shape(operand)
-  dtype = shape.element_type().type
-  if np.issubdtype(dtype, np.complexfloating):
-    raise NotImplementedError("eigh is not implemented on TPU for complex inputs.")
-  return eigh_translation_rule(c, operand, lower)
-
 def eigh_jvp_rule(primals, tangents, lower):
   # Derivative for eigh in the simplest case of distinct eigenvalues.
   # This is classic nondegenerate perurbation theory, but also see
@@ -570,8 +559,6 @@ if rocsolver is not None:
   xla.backend_specific_translations['gpu'][eigh_p] = partial(
     _eigh_cpu_gpu_translation_rule, rocsolver.syevd)
 
-xla.backend_specific_translations['tpu'][eigh_p] = _eigh_tpu_translation_rule
-
 
 triangular_solve_dtype_rule = partial(
     naryop_dtype_rule, _input_dtype, (_float | _complex, _float | _complex),
@@ -1002,7 +989,7 @@ def _lu_batching_rule(batched_args, batch_dims):
   x = batching.moveaxis(x, bd, 0)
   return lu_p.bind(x), (0, 0, 0)
 
-def _lu_cpu_gpu_translation_rule(getrf_impl, c, operand):
+def _lu_cpu_gpu_translation_rule(getrf_impl, c, operand, backend):
   shape = c.get_shape(operand)
   batch_dims = shape.dimensions()[:-2]
   m = shape.dimensions()[-2]
@@ -1013,7 +1000,7 @@ def _lu_cpu_gpu_translation_rule(getrf_impl, c, operand):
   lu = _broadcasting_select(c, xops.Reshape(ok, batch_dims + (1, 1)), lu,
                             _nan_like(c, lu))
   perm = xla.lower_fun(lambda x: lu_pivots_to_permutation(x, m),
-                       multiple_results=False)(c, pivot)
+                       multiple_results=False, backend=backend)(c, pivot)
   return xops.Tuple(c, [lu, pivot, perm])
 
 
@@ -1034,15 +1021,15 @@ ad.primitive_jvps[lu_p] = _lu_jvp_rule
 batching.primitive_batchers[lu_p] = _lu_batching_rule
 
 xla.backend_specific_translations['cpu'][lu_p] = partial(
-  _lu_cpu_gpu_translation_rule, lapack.getrf)
+  _lu_cpu_gpu_translation_rule, lapack.getrf, backend='cpu')
 
 if cusolver is not None:
   xla.backend_specific_translations['gpu'][lu_p] = partial(
-    _lu_cpu_gpu_translation_rule, cusolver.getrf)
+    _lu_cpu_gpu_translation_rule, cusolver.getrf, backend='gpu')
 
 if rocsolver is not None:
   xla.backend_specific_translations['gpu'][lu_p] = partial(
-    _lu_cpu_gpu_translation_rule, rocsolver.getrf)
+    _lu_cpu_gpu_translation_rule, rocsolver.getrf, backend='gpu')
 
 xla.backend_specific_translations['tpu'][lu_p] = _lu_tpu_translation_rule
 
