@@ -25,7 +25,7 @@ from jax import numpy as jnp
 from jax._src import dtypes
 from jax.core import NamedShape
 from jax._src.api import jit, vmap
-from jax._src.numpy.lax_numpy import _constant_like, _convert_and_clip_integer, asarray
+from jax._src.numpy.lax_numpy import _constant_like, _convert_and_clip_integer, _check_arraylike
 from jax.lib import xla_bridge
 from jax.lib import xla_client
 from jax.lib import cuda_prng
@@ -86,14 +86,6 @@ def _is_prng_key(key: jnp.ndarray) -> bool:
 ### utilities
 
 
-# TODO(mattjj,jakevdp): add more info to error message, use this utility more
-def _asarray(x):
-  """A more restrictive jnp.asarray, only accepts JAX arrays and np.ndarrays."""
-  if not isinstance(x, (np.ndarray, jnp.ndarray)):
-    raise TypeError(f"Function requires array input, got {x} of type {type(x)}.")
-  return jnp.asarray(x)
-
-
 def _make_rotate_left(dtype):
   if not jnp.issubdtype(dtype, np.integer):
     raise TypeError("_rotate_left only accepts integer dtypes.")
@@ -143,7 +135,7 @@ def rolled_loop_step(i, state):
   x, ks, rotations = state
   for r in rotations[0]:
     x = apply_round(x, r)
-  new_x = [x[0] + ks[0], x[1] + ks[1] + asarray(i + 1, dtype=np.uint32)]
+  new_x = [x[0] + ks[0], x[1] + ks[1] + jnp.asarray(i + 1, dtype=np.uint32)]
   return new_x, rotate_list(ks), rotate_list(rotations)
 
 def _threefry2x32_lowering(key1, key2, x1, x2, use_rolled_loops=True):
@@ -445,8 +437,9 @@ def _randint(key, shape, minval, maxval, dtype):
   if not jnp.issubdtype(dtype, np.integer):
     raise TypeError(f"randint only accepts integer dtypes, got {dtype}")
 
-  minval = _asarray(minval)
-  maxval = _asarray(maxval)
+  _check_arraylike("randint", minval, maxval)
+  minval = jnp.asarray(minval)
+  maxval = jnp.asarray(maxval)
   if not jnp.issubdtype(minval.dtype, np.integer):
     minval = minval.astype(int)
   if not jnp.issubdtype(maxval.dtype, np.integer):
@@ -501,7 +494,7 @@ def _randint(key, shape, minval, maxval, dtype):
   return lax.add(minval, lax.convert_element_type(random_offset, dtype))
 
 
-def shuffle(key: jnp.ndarray, x: jnp.ndarray, axis: int = 0) -> jnp.ndarray:
+def shuffle(key: jnp.ndarray, x: Array, axis: int = 0) -> jnp.ndarray:
   """Shuffle the elements of an array uniformly at random along an axis.
 
   Args:
@@ -518,7 +511,7 @@ def shuffle(key: jnp.ndarray, x: jnp.ndarray, axis: int = 0) -> jnp.ndarray:
   return _shuffle(key, x, axis)  # type: ignore
 
 
-def permutation(key: jnp.ndarray, x: IntegerArray) -> jnp.ndarray:
+def permutation(key: jnp.ndarray, x: Array) -> jnp.ndarray:
   """
   Permute elements of an array along its first axis or return a permuted range.
 
@@ -603,10 +596,11 @@ def choice(key: jnp.ndarray,
                     f"got {shape}")
   if np.ndim(a) not in [0, 1]:
     raise ValueError("a must be an integer or 1-dimensional")
+  _check_arraylike("choice", a)
   if np.ndim(a) == 0:
-    a = int(a)
+    a = core.concrete_or_error(int, a, "The error occurred in jax.random.choice()")
   else:
-    a = _asarray(a)
+    a = jnp.asarray(a)
   n_inputs = int(a) if np.ndim(a) == 0 else len(a)  # type: ignore[arg-type]
   n_draws = prod(shape)
   if n_draws == 0:
@@ -666,9 +660,9 @@ def _normal(key, shape, dtype) -> jnp.ndarray:
     sqrt2 = np.array(np.sqrt(2), dtype)
 
     key_re, key_im = split(key)
-    dtype = dtypes.dtype_real(dtype)
-    _re = _normal_real(key_re, shape, dtype)
-    _im = _normal_real(key_im, shape, dtype)
+    real_dtype = np.array(0, dtype).real.dtype
+    _re = _normal_real(key_re, shape, real_dtype)
+    _im = _normal_real(key_im, shape, real_dtype)
     return (_re + 1j * _im) / sqrt2
   else:
     return _normal_real(key, shape, dtype) # type: ignore
@@ -1019,7 +1013,7 @@ def _gamma_one(key, alpha):
   alpha = lax.select(lax.ge(alpha, one), alpha, lax.add(alpha, one))
 
   d = lax.sub(alpha, one_over_three)
-  c = lax.div(one_over_three, lax.pow(d, one_over_two))
+  c = lax.div(one_over_three, lax.sqrt(d))
 
   def _cond_fn(kXVU):
     _, X, V, U = kXVU
@@ -1281,7 +1275,7 @@ def _gumbel(key, shape, dtype):
 
 
 def categorical(key: jnp.ndarray,
-                logits: jnp.ndarray,
+                logits: RealArray,
                 axis: int = -1,
                 shape: Optional[Sequence[int]] = None) -> jnp.ndarray:
   """Sample random values from categorical distributions.
