@@ -15,7 +15,7 @@
 
 import operator
 from operator import attrgetter
-from contextlib import contextmanager, suppress
+from contextlib import contextmanager
 from collections import namedtuple
 from functools import total_ordering
 import itertools as it
@@ -450,13 +450,10 @@ def escaped_tracer_error(tracer, detail=None):
       msg += (f'When the tracer was created, the final {num_frames} stack '
               'frames (most recent last) excluding JAX-internal frames were:\n'
               f'{source_info_util.summarize(line_info, num_frames=num_frames)}')
-  try:
-    fun_source_info = tracer._trace.main.source_info
-  except AttributeError:
-    pass
-  else:
+  dbg = getattr(tracer._trace.main, 'debug_info', None)
+  if dbg is not None:
     msg += ('\nThe function being traced when the tracer leaked was '
-            f'{fun_source_info}.')
+            f'{dbg.func_src_info} traced for {dbg.traced_for}.')
   msg += ('\nTo catch the leak earlier, try setting the environment variable '
           'JAX_CHECK_TRACER_LEAKS or using the `jax.checking_leaks` context '
           'manager.')
@@ -611,11 +608,13 @@ class EvalTrace(Trace):
 
   def process_custom_jvp_call(self, primitive, fun, jvp, tracers):
     del primitive, jvp  # Unused.
-    return fun.call_wrapped(*tracers)
+    with new_sublevel():
+      return fun.call_wrapped(*tracers)
 
   def process_custom_vjp_call(self, primitive, fun, fwd, bwd, tracers, out_trees):
     del primitive, fwd, bwd, out_trees  # Unused.
-    return fun.call_wrapped(*tracers)
+    with new_sublevel():
+      return fun.call_wrapped(*tracers)
 
 
 class MainTrace:
@@ -806,11 +805,6 @@ def new_sublevel() -> Generator[None, None, None]:
     if t() is not None:
       raise Exception(f'Leaked sublevel {t()}.')
 
-def maybe_new_sublevel(trace):
-  # dynamic traces run the WrappedFun, so we raise the sublevel for them
-  dynamic = thread_local_state.trace_state.trace_stack.dynamic
-  return new_sublevel() if trace.main is dynamic else suppress()
-
 def full_lower(val):
   if isinstance(val, Tracer):
     return val.full_lower()
@@ -1266,11 +1260,11 @@ class DimensionHandler:
     """
     return d1 - d2
 
-  def divide_shape_sizes(self, s1: Shape, s2: Shape) -> int:
-    """Computes the division of the sizes of the shapes.
+  def divide_shape_sizes(self, s1: Shape, s2: Shape) -> DimSize:
+    """Computes integer "i" such that i  * size(s2) == size(s1).
 
-    Raise InconclusiveDimensionOperation if the result is different in different
-    contexts, or if the division is not even.
+    Raise InconclusiveDimensionOperation if there is no such integer for all
+    contexts,
     """
     sz1 = int(np.prod(s1))
     sz2 = int(np.prod(s2))
@@ -1353,7 +1347,9 @@ def diff_dim(d1: DimSize, d2: DimSize) -> DimSize:
 def diff_shape(s1: Shape, s2: Shape) -> Shape:
   return tuple(map(diff_dim, s1, s2))
 
-def divide_shape_sizes(s1: Shape, s2: Shape) -> int:
+def divide_shape_sizes(s1: Shape, s2: Shape) -> DimSize:
+  """Returns an integer "i" s.t., i * size(s2) == size(s1).
+  Raises if there is no such integer."""
   s1 = s1 or (1,)
   s2 = s2 or (1,)
   handler, ds = _dim_handler_and_canonical(*s1, *s2)
@@ -1552,8 +1548,7 @@ def call_bind(primitive: Union['CallPrimitive', 'MapPrimitive'],
       fun, primitive, top_trace and top_trace.level,
       params_tuple, out_axes_transforms)
   tracers = map(top_trace.full_raise, args)
-  with maybe_new_sublevel(top_trace):
-    outs = primitive.process(top_trace, fun, tracers, params)
+  outs = primitive.process(top_trace, fun, tracers, params)
   return map(full_lower, apply_todos(env_trace_todo(), outs))
 
 
@@ -1572,7 +1567,8 @@ class CallPrimitive(Primitive):
 
 def call_impl(f: lu.WrappedFun, *args, **params):
   del params  # params parameterize the call primitive, not the function
-  return f.call_wrapped(*args)
+  with new_sublevel():
+    return f.call_wrapped(*args)
 
 call_p = CallPrimitive('call')
 call = call_p.bind
@@ -1697,9 +1693,11 @@ def used_axis_names(primitive: Primitive, params: ParamDict) -> Set[AxisName]:
   subst_axis_names(primitive, params, register_name)
   return axis_names
 
-def subst_axis_names(primitive: Primitive, params: ParamDict, subst: AxisSubst) -> ParamDict:
+def subst_axis_names(primitive: Primitive, params: ParamDict, subst: AxisSubst, traverse: bool = True) -> ParamDict:
   if primitive in axis_substitution_rules:
-    return axis_substitution_rules[primitive](params, subst)
+    return axis_substitution_rules[primitive](params, subst, traverse)
+  if not traverse:
+    return params
   # Default implementation: substitute names in all jaxpr parameters
   if isinstance(primitive, MapPrimitive):
     def shadowed_subst(name):
@@ -1760,7 +1758,7 @@ def subst_axis_names_jaxpr(jaxpr: Union[Jaxpr, ClosedJaxpr], subst: AxisSubst):
     return ClosedJaxpr(new_jaxpr, consts)
   return new_jaxpr
 
-axis_substitution_rules: Dict[Primitive, Callable[[ParamDict, AxisSubst], ParamDict]] = {}
+axis_substitution_rules: Dict[Primitive, Callable[[ParamDict, AxisSubst, bool], ParamDict]] = {}
 
 # ------------------- AxisPrimitive -------------------
 # Primitives that store axis names in params and want those axis names to
