@@ -22,11 +22,14 @@ from absl.testing import parameterized
 
 import jax
 from jax import dtypes
+from jax import lax
 from jax import numpy as jnp
 from jax import test_util as jtu
 from jax.config import config
 from jax.experimental import jax2tf
 from jax.experimental.jax2tf.tests import tf_test_util
+from jax._src import source_info_util
+
 import numpy as np
 import tensorflow as tf  # type: ignore[import]
 
@@ -37,7 +40,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
 
   def test_basics(self):
     f_jax = lambda x: jnp.sin(jnp.cos(x))
-    _, res_tf = self.ConvertAndCompare(f_jax, jnp.float_(0.7))
+    _, res_tf = self.ConvertAndCompare(f_jax, 0.7)
 
   def test_input_output_naming(self):
     @jax2tf.convert
@@ -56,20 +59,20 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     cf = u.get_concrete_function([1., 2., 3.], 4.)
     g = cf.graph
     g.get_operation_by_name("jax2tf_arg_0")
-    g.get_operation_by_name("jax2tf_arg_0_1")
-    g.get_operation_by_name("jax2tf_arg_0_2")
     g.get_operation_by_name("jax2tf_arg_1")
+    g.get_operation_by_name("jax2tf_arg_2")
+    g.get_operation_by_name("jax2tf_arg_3")
     g.get_operation_by_name("jax2tf_out")
     g.get_operation_by_name("jax2tf_out_1")
     g.get_operation_by_name("jax2tf_out_2")
     with self.assertRaises(KeyError):
-      g.get_operation_by_name("jax2tf_arg_2")
+      g.get_operation_by_name("jax2tf_arg_4")
     with self.assertRaises(KeyError):
       g.get_operation_by_name("jax2tf_out_3")
     g.get_operation_by_name("jax2tf_vjp/jax2tf_arg_0")
     g.get_operation_by_name("jax2tf_vjp/jax2tf_arg_1")
-    g.get_operation_by_name("jax2tf_vjp/jax2tf_arg_1_1")
-    g.get_operation_by_name("jax2tf_vjp/jax2tf_arg_1_2")
+    g.get_operation_by_name("jax2tf_vjp/jax2tf_arg_2")
+    g.get_operation_by_name("jax2tf_vjp/jax2tf_arg_3")
     g.get_operation_by_name("jax2tf_vjp/jax2tf_out")
     g.get_operation_by_name("jax2tf_vjp/jax2tf_out_1")
     g.get_operation_by_name("jax2tf_vjp/jax2tf_out_2")
@@ -79,21 +82,21 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     # Take and return pytrees
     def f_jax(x: Tuple[float, Dict[str, float]]) -> Tuple[float, Dict[str, float]]:
       x_a, x_dict = x
-      return x_a * 2., {k : v * 3. for k, v in x_dict.items()}
+      return x_a * 2., {k: v * 3. for k, v in x_dict.items()}
 
-    x = (jnp.float_(.7), {"a": jnp.float_(.8), "b": jnp.float_(.9)})
+    x = (.7, {"a": .8, "b": .9})
     self.ConvertAndCompare(f_jax, x)
 
   def test_variable_input(self):
     f_jax = lambda x: jnp.sin(jnp.cos(x))
     f_tf = jax2tf.convert(f_jax)
-    v = tf.Variable(0.7, dtype=dtypes.canonicalize_dtype(jnp.float_))
+    v = tf.Variable(0.7, dtype=jax2tf.dtype_of_val(0.7))
     self.assertIsInstance(f_tf(v), tf.Tensor)
     self.assertAllClose(f_jax(0.7), f_tf(v))
 
   def test_jit(self):
     f_jax = jax.jit(lambda x: jnp.sin(jnp.cos(x)))
-    self.ConvertAndCompare(f_jax, jnp.float_(0.7))
+    self.ConvertAndCompare(f_jax, 0.7)
 
   def test_nested_jit(self):
     f_jax = jax.jit(lambda x: jnp.sin(jax.jit(jnp.cos)(x)))
@@ -131,9 +134,11 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     self.assertEqual(f_tf(1., 2.).dtype, tf.bfloat16)
 
   @parameterized.named_parameters(jtu.cases_from_list(
-    dict(testcase_name=f"_dtype={dtype.__name__}",
-         dtype=dtype)
-    for dtype in [np.int64, np.float64]))
+    dict(testcase_name=f"_dtype={dtype.__name__}_function={with_function}",
+         dtype=dtype,
+         with_function=with_function)
+    for dtype in [np.int64, np.float64]
+    for with_function in [True, False]))
   def test_converts_64bit(self, dtype=np.int64, with_function=False):
     if not config.jax_enable_x64:
       self.skipTest("requires x64 mode")
@@ -149,14 +154,40 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     self.assertAllClose(jnp.sin(big_const),
                         f_conv(tf.constant(big_const)))
 
+  def test_64bit_behavior_enable_x64(self):
+    if not config.jax_enable_x64:
+      self.skipTest("requires x64 mode")
+
+    # JAX and TF have different default float types if JAX_ENABLE_X64=1
+    self.assertEqual(tf.math.sin(0.7).dtype, tf.float32)
+    self.assertEqual(jnp.sin(0.7).dtype, jnp.float64)
+
+    # jax2tf.convert has the same behavior as JAX
+    self.assertEqual(jax2tf.convert(jnp.sin)(0.7).dtype, tf.float64)
+
+  def test_64bit_behavior_not_enable_x64(self):
+    if config.jax_enable_x64:
+      self.skipTest("requires not x64 mode")
+
+    # JAX and TF have same default float types if JAX_ENABLE_X64=1
+    self.assertEqual(tf.math.sin(0.7).dtype, tf.float32)
+    self.assertEqual(jnp.sin(0.7).dtype, jnp.float32)
+
+    # Except that JAX forces values to 32-bit
+    self.assertEqual(jnp.sin(np.float64(0.7)).dtype, jnp.float32)
+
+    # jax2tf.convert has the same behavior as JAX
+    self.assertEqual(jax2tf.convert(jnp.sin)(0.7).dtype, tf.float32)
+    self.assertEqual(jax2tf.convert(jnp.sin)(np.float64(0.7)).dtype, tf.float32)
+
   def test_function(self):
     f_jax = jax.jit(lambda x: jnp.sin(jnp.cos(x)))
-    self.ConvertAndCompare(f_jax, jnp.float_(0.7))
+    self.ConvertAndCompare(f_jax, 0.7)
 
   @parameterized.named_parameters(jtu.cases_from_list(
-    dict(testcase_name=f"function={with_function}",
-         with_function=with_function)
-    for with_function in [False, True]))
+      dict(testcase_name=f"function={with_function}",
+           with_function=with_function)
+      for with_function in [False, True]))
   def test_gradients_disabled(self, with_function=False):
     f_tf = jax2tf.convert(jnp.tan, with_gradient=False)
     if with_function:
@@ -182,8 +213,8 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     f_tf = jax2tf.convert(f, with_gradient=True)
     if with_function:
       f_tf = tf.function(f_tf, autograph=False)
-    default_float_type = dtypes.canonicalize_dtype(jnp.float_)
-    x = tf.Variable(4., dtype=default_float_type)
+    default_float_type = jax2tf.dtype_of_val(4.)
+    x = tf.Variable(4., dtype=jax2tf.dtype_of_val(4.))
     y = tf.Variable(5., dtype=default_float_type)
     with tf.GradientTape(persistent=True) as tape:
       u, v = f_tf(x, y)
@@ -205,7 +236,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     f_tf = jax2tf.convert(f, with_gradient=True)
     if with_function:
       f_tf = tf.function(f_tf, autograph=False)
-    default_float_dtype = dtypes.canonicalize_dtype(jnp.float_)
+    default_float_dtype = jax2tf.dtype_of_val(4.)
     x = tf.Variable(4., dtype=default_float_dtype)
     y = tf.Variable(5., dtype=default_float_dtype)
     with tf.GradientTape(persistent=True) as tape:
@@ -217,9 +248,9 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     self.assertAllClose(4., tape.gradient(uv["two"], y))
 
   @parameterized.named_parameters(jtu.cases_from_list(
-    dict(testcase_name=f"function={with_function}",
-         with_function=with_function)
-    for with_function in [False, True]))
+      dict(testcase_name=f"function={with_function}",
+           with_function=with_function)
+      for with_function in [False, True]))
   def test_gradients_with_custom_jvp(self, with_function=True):
     """Check gradients, for a function with custom JVP."""
     @jax.custom_jvp
@@ -241,8 +272,8 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     f_tf = jax2tf.convert(f, with_gradient=True)
     if with_function:
       f_tf = tf.function(f_tf, autograph=False)
-    self.assertAllClose(4. * 4., f_tf(jnp.float_(4.)))
-    x = tf.Variable(4., dtype=dtypes.canonicalize_dtype(jnp.float_))
+    self.assertAllClose(4. * 4., f_tf(4.))
+    x = tf.Variable(4., dtype=jax2tf.dtype_of_val(4.))
     with tf.GradientTape() as tape:
       tape.watch(x)
       y = f_tf(x)
@@ -251,9 +282,9 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     self.assertAllClose(3. * 4., tape.gradient(y, x))
 
   @parameterized.named_parameters(jtu.cases_from_list(
-    dict(testcase_name=f"function={with_function}",
-         with_function=with_function)
-    for with_function in [False, True]))
+      dict(testcase_name=f"function={with_function}",
+           with_function=with_function)
+      for with_function in [False, True]))
   def test_gradients_with_custom_vjp(self, with_function=True):
     """Check gradients, for a function with custom VJP."""
     @jax.custom_vjp
@@ -275,8 +306,8 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     f_tf = jax2tf.convert(f, with_gradient=True)
     if with_function:
       f_tf = tf.function(f_tf, autograph=False)
-    self.assertAllClose(4. * 4., f_tf(jnp.float_(4.)))
-    x = tf.Variable(4., dtype=dtypes.canonicalize_dtype(jnp.float_))
+    self.assertAllClose(4. * 4., f_tf(4.))
+    x = tf.Variable(4., dtype=jax2tf.dtype_of_val(4.))
     with tf.GradientTape() as tape:
       tape.watch(x)
       y = f_tf(x)
@@ -292,7 +323,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     def g(x):  # x: f32
       return 2. * f(3 * x.astype("int32"), x * 4.)
 
-    x = np.float_(2.)
+    x = 2.
     grad_g = jax.grad(g)
     self.ConvertAndCompare(grad_g, x)
 
@@ -348,6 +379,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
 
   def test_custom_jvp(self):
     """Conversion of function with custom JVP"""
+
     @jax.custom_jvp
     def f(x):
       return x * x
@@ -360,7 +392,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
       tangent_out = 3. * x * x_dot
       return primal_out, tangent_out
 
-    arg = jnp.float_(0.7)
+    arg = 0.7
     self.TransformConvertAndCompare(f, arg, None)
     self.TransformConvertAndCompare(f, arg, "jvp")
     self.TransformConvertAndCompare(f, arg, "vmap")
@@ -370,6 +402,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
 
   def test_custom_vjp(self):
     """Conversion of function with custom VJP"""
+
     @jax.custom_vjp
     def f(x):
       return x * x
@@ -377,12 +410,13 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     # f_fwd: a -> (b, residual)
     def f_fwd(x):
       return f(x), 3. * x
+
     # f_bwd: (residual, CT b) -> [CT a]
     def f_bwd(residual, ct_b):
       return residual * ct_b,
 
     f.defvjp(f_fwd, f_bwd)
-    arg = jnp.float_(0.7)
+    arg = 0.7
     self.TransformConvertAndCompare(f, arg, None)
     self.TransformConvertAndCompare(f, arg, "vmap")
     self.TransformConvertAndCompare(f, arg, "grad")
@@ -402,7 +436,6 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     self.TransformConvertAndCompare(f, arg, "grad")
     # TODO: check that the TF code also computes "sin" 5 times
 
-
   def test_remat_free_var(self):
     def f(x):
       y = 2 * x
@@ -412,7 +445,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
         return y
 
       return g()
-    arg = jnp.float_(3.)
+    arg = 3.
     self.TransformConvertAndCompare(f, arg, None)
     self.TransformConvertAndCompare(f, arg, "grad")
 
@@ -493,7 +526,10 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
       log.append(y.name)
       return x * x
 
-    jax2tf.convert(my_test_function)(2)
+    def caller(x):
+      return my_test_function(jnp.sin(x))
+
+    jax2tf.convert(caller)(2.)
     self.assertIn("my_test_function/foo", log[0])
 
   def test_bfloat16_constant(self):
@@ -516,6 +552,232 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
         tf_fn_array(np.array([3, 4, 5])), np.array([4.5, 10, 17.5],
                                                    jnp.bfloat16))
 
+  def test_kwargs(self):
+    # Re: https://github.com/google/jax/issues/6791
+    def f_jax(*, x):
+      return jnp.sum(x)
+    f_tf = jax2tf.convert(f_jax)
+    self.assertAllClose(
+      f_tf(x=np.zeros(3, dtype=np.float32)),  # Call with kwargs.
+      np.zeros((), dtype=np.float32))
+
+  def test_enable_xla(self):
+    # Tests that enable_xla flag is properly scoped to a conversion.
+    def fun(x):
+      # Can be converted only if enable_xla is on, due to negative padding.
+      return lax.pad(x, np.float32(0), [(-1, 0, 0), (0, 0, 0)])
+
+    tf_fun_with_xla = jax2tf.convert(fun, enable_xla=True)
+    tf_fun_without_xla = jax2tf.convert(fun, enable_xla=False)
+    x = np.ones((2, 3), dtype=np.float32)
+
+    self.assertAllClose(fun(x), tf_fun_with_xla(x))
+    with self.assertRaisesRegex(NotImplementedError,
+                                "Call to pad cannot be converted with enable_xla=False"):
+      tf_fun_without_xla(x)
+
+    # Now in reverse order
+    def fun2(x):
+      # Can be converted only if enable_xla is on, due to negative padding.
+      return lax.pad(x, np.float32(0), [(-1, 0, 0), (0, 0, 0)])
+
+    tf_fun2_without_xla = jax2tf.convert(fun2, enable_xla=False)
+    tf_fun2_with_xla = jax2tf.convert(fun2, enable_xla=True)
+
+    with self.assertRaisesRegex(NotImplementedError,
+                                "Call to pad cannot be converted with enable_xla=False"):
+      tf_fun2_without_xla(x)
+    self.assertAllClose(fun(x), tf_fun2_with_xla(x))
+
+  def test_device_array_arg(self):
+    self.ConvertAndCompare(jnp.sin, jnp.zeros((2, 3), jnp.float32))
+
+  def test_randint(self):
+      def randint():
+        return jax.random.randint(
+            jax.random.PRNGKey(42), shape=(), minval=0, maxval=1)
+
+      self.ConvertAndCompare(randint)
+
+  def test_op_metadata_simple(self):
+    self.skipTest("include_xla_op_metadata not yet enabled")
+    # A simple example
+    # The user_frame is used to compute line numbers for ops in the test.
+    user_frame = source_info_util.user_frame(source_info_util.current())
+    def f_simple(x):
+      return jnp.sin(x)
+
+    x = np.ones((2, 3), np.float32)
+    self.CheckOpMetadata(
+        f_simple, x,
+        [tf_test_util.OpMetadataGraph(tf_type="Sin",
+                                      source_file=__file__,
+                                      source_line=user_frame.line_num + 2,
+                                      op_name="jax2tf(f_simple)/sin",
+                                      op_type="sin")
+         ]
+    )
+
+  def test_op_metadata_sub_jit(self):
+    self.skipTest("include_xla_op_metadata not yet enabled")
+    # Calling a jitted-function
+    # The user_frame is used to compute line numbers for ops in the test.
+    user_frame = source_info_util.user_frame(source_info_util.current())
+    def f_callee(x):
+      return jnp.cos(x)
+    def f_caller(x):
+      y = jnp.tanh(x)
+      z = jax.jit(f_callee)(y)
+      return jnp.sin(z)
+
+    x = np.ones((2, 3), np.float32)
+
+    self.CheckOpMetadata(
+        f_caller, x,
+        [tf_test_util.OpMetadataGraph(tf_type="Tanh",
+                                      source_file=__file__,
+                                      source_line=user_frame.line_num + 4,
+                                      op_name="jax2tf(f_caller)/tanh",
+                                      op_type="tanh"),
+         tf_test_util.OpMetadataGraph(tf_type="Cos",
+                                      source_file=__file__,
+                                      source_line=user_frame.line_num + 2,
+                                      op_name="jax2tf(f_caller)/jit(f_callee)/cos",
+                                      op_type="cos"),
+         tf_test_util.OpMetadataGraph(tf_type="Sin",
+                                      source_file=__file__,
+                                      source_line=user_frame.line_num + 6,
+                                      op_name="jax2tf(f_caller)/sin",
+                                      op_type="sin"),
+         ]
+    )
+
+  def test_op_metadata_named(self):
+    self.skipTest("include_xla_op_metadata not yet enabled")
+    # Calling a jax.named_call
+    # The user_frame is used to compute line numbers for ops in the test.
+    user_frame = source_info_util.user_frame(source_info_util.current())
+    def f_callee(x):
+      return jnp.cos(x)
+    def f_caller(x):
+      y = jnp.tanh(x)
+      z = jax.named_call(f_callee, name="callee")(y)
+      return jnp.sin(z)
+
+    x = np.ones((2, 3), np.float32)
+
+    self.CheckOpMetadata(
+        f_caller, x,
+        [tf_test_util.OpMetadataGraph(tf_type="Tanh",
+                                      source_file=__file__,
+                                      source_line=user_frame.line_num + 4,
+                                      op_name="jax2tf(f_caller)/tanh",
+                                      op_type="tanh"),
+         tf_test_util.OpMetadataGraph(tf_type="Cos",
+                                      source_file=__file__,
+                                      source_line=user_frame.line_num + 2,
+                                      op_name="jax2tf(f_caller)/named(callee)/cos",
+                                      op_type="cos"),
+         tf_test_util.OpMetadataGraph(tf_type="Sin",
+                                      source_file=__file__,
+                                      source_line=user_frame.line_num + 6,
+                                      op_name="jax2tf(f_caller)/sin",
+                                      op_type="sin"),
+         ]
+    )
+
+  def test_op_metadata_while_and_cond(self):
+    self.skipTest("include_xla_op_metadata not yet enabled")
+    # An example with while and cond
+    # The user_frame is used to compute line numbers for ops in the test.
+    user_frame = source_info_util.user_frame(source_info_util.current())
+    def f_while_cond(x):
+      def body_fun(i_acc):
+        i, acc = i_acc
+        return (i + 1,
+                (jnp.cos(acc) +
+                 lax.cond(jnp.mod(i, 2) == 0,
+                          lambda acc: jnp.sin(acc),
+                          lambda acc: acc,
+                          acc)))
+
+      _, acc = lax.while_loop(
+          lambda i_acc: i_acc[0] <= 5,
+          body_fun, (0, x))
+      return acc
+
+    x = np.ones((2, 3), np.float32)
+    self.CheckOpMetadata(
+        f_while_cond, x,
+        [tf_test_util.OpMetadataGraph(tf_type="Cos",
+                                      source_file=__file__,
+                                      source_line=user_frame.line_num + 5,
+                                      op_name="jax2tf(f_while_cond)/while/body/cos",
+                                      op_type="cos"),
+         tf_test_util.OpMetadataGraph(tf_type="Sin",
+                                      source_file=__file__,
+                                      source_line=user_frame.line_num + 7,
+                                      op_name="jax2tf(f_while_cond)/while/body/branch_1_fun/sin",
+                                      op_type="sin"),
+         tf_test_util.OpMetadataGraph(tf_type="FloorMod",
+                                      source_file=__file__,
+                                      source_line=user_frame.line_num + 6,
+                                      op_name="jax2tf(f_while_cond)/while/body/rem",
+                                      op_type="rem"),
+         ]
+    )
+
+  def test_op_metadata_batched_while(self):
+    self.skipTest("include_xla_op_metadata not yet enabled")
+    # An example with while and cond
+    # The user_frame is used to compute line numbers for ops in the test.
+    user_frame = source_info_util.user_frame(source_info_util.current())
+    @jax.vmap
+    def f_while(x):
+      def body_fun(carry):
+        new_carry = jnp.sin(carry)  # We look for "sin" in the graph
+        return new_carry
+
+      _, carry = lax.while_loop(
+          lambda carry: jnp.all(carry <= x),  # We look for "le" in the graph
+          body_fun, x)
+      return carry
+
+    shape = (3, 2)
+    x = np.arange(np.prod(shape), dtype=np.float32).reshape(shape)
+
+    jax_comp = jax.xla_computation(f_while)(x)
+    backend = jax.lib.xla_bridge.get_backend()
+    modules = backend.compile(jax_comp).hlo_modules()
+    jax_opt_hlo = modules[0].to_string()
+    print(f"JAX OPT HLO = {jax_opt_hlo}")
+
+    self.CheckOpMetadata(
+        f_while, x,
+        [tf_test_util.OpMetadataGraph(tf_type="Sin",
+                                      source_file=__file__,
+                                      source_line=user_frame.line_num + 4,
+                                      op_name="jax2tf(f_while)/while/body/sin",
+                                      op_type="sin"),
+         tf_test_util.OpMetadataGraph(tf_type="LessEqual",
+                                      source_file=__file__,
+                                      source_line=user_frame.line_num + 8,
+                                      op_name="jax2tf(f_while)/while/body_pred/le",
+                                      op_type="le"),
+         ]
+    )
+
+  def test_op_metadata_disabled(self):
+    self.skipTest("include_xla_op_metadata not yet enabled")
+    def f_simple(x):
+      return jnp.sin(x)
+
+    x = np.ones((2, 3), np.float32)
+    self.CheckOpMetadata(
+        f_simple, x,
+        [],
+        include_xla_op_metadata=False
+    )
 
 if __name__ == "__main__":
   absltest.main(testLoader=jtu.JaxTestLoader())
