@@ -99,8 +99,7 @@ class JaxPrimitiveTest(tf_test_util.JaxToTfTestCase):
   # If you want to run this test for only one harness, add parameter
   # `one_containing="foo"` to parameterized below.
   @primitive_harness.parameterized(
-      primitive_harness.all_harnesses, include_jax_unimpl=False
-      )
+      primitive_harness.all_harnesses, include_jax_unimpl=False)
   @jtu.ignore_warning(
       category=UserWarning, message="Using reduced precision for gradient.*")
   def test_prim(self, harness: primitive_harness.Harness):
@@ -110,7 +109,9 @@ class JaxPrimitiveTest(tf_test_util.JaxToTfTestCase):
                                                   dtype=harness.dtype), limitations))
     func_jax = harness.dyn_fun
     args = harness.dyn_args_maker(self.rng())
-    self.ConvertAndCompare(func_jax, *args, limitations=limitations)
+    enable_xla = harness.params.get("enable_xla", True)
+    self.ConvertAndCompare(func_jax, *args, limitations=limitations,
+                           enable_xla=enable_xla)
 
   def test_primitive_coverage(self):
     """Fail if there are JAX primitives that are not implemented."""
@@ -259,8 +260,7 @@ class JaxPrimitiveTest(tf_test_util.JaxToTfTestCase):
       return lax.pad(x, np.float32(0), [(-1, 0, 0), (0, 0, 0)])
 
     with self.assertRaisesRegex(
-        NotImplementedError, "Call to pad can only be converted through "
-                             "TFXLA, but XLA is disabled"):
+        NotImplementedError, "Call to pad cannot be converted with enable_xla=False."):
       self.ConvertAndCompare(
           fun, np.ones((2, 3), dtype=np.float32), enable_xla=False)
 
