@@ -13,6 +13,7 @@
 # limitations under the License.
 
 import contextlib
+import dataclasses
 import logging
 
 from typing import Any, Callable, List, Optional, Sequence
@@ -30,27 +31,16 @@ from jax.interpreters import masking
 from jax._src import util
 import numpy as np
 import tensorflow as tf  # type: ignore[import]
+from tensorflow.compiler.xla import xla_data_pb2  # type: ignore[import]
 
 DType = Any
 
-def _make_tf_args(args):
-  def _convert_to_tensor(v):
-    if hasattr(v, "dtype"):
-      tf.convert_to_tensor(v)
-    return v
-
-  return tf.nest.map_structure(_convert_to_tensor, args)
-
-
 def _make_tf_input_signature(*tf_args) -> List[tf.TensorSpec]:
   # tf_args can be PyTrees
-  def _make_one_arg_signature(tf_arg):
-    if np.isscalar(tf_arg):
-      tf_arg = np.array(tf_arg)
-    return tf.TensorSpec(np.shape(tf_arg), tf_arg.dtype)
-
-  return tf.nest.map_structure(_make_one_arg_signature, list(tf_args))
+  def _make_one_array_signature(tf_arg):
+    return tf.TensorSpec(np.shape(tf_arg), jax2tf.dtype_of_val(tf_arg))
 
+  return tf.nest.map_structure(_make_one_array_signature, list(tf_args))
 
 def _run_tf_function(func_tf: Callable, *tf_args, mode: str):
   if mode == "eager":
@@ -74,6 +64,16 @@ def _run_tf_function(func_tf: Callable, *tf_args, mode: str):
         f"Expected 'eager', 'graph', or 'compiled' for mode: got '{mode}'")
 
 
+## Helper functions for matching OpMetadata in TF graphs
+@dataclasses.dataclass(order=True, frozen=True)
+class OpMetadataGraph:
+  tf_type: str  # The standard Tf.Operation.type
+  op_type: str  # The rest are OpMetadata fields from _Xla... attributes
+  op_name: str
+  source_file: str
+  source_line: str
+
+
 class JaxToTfTestCase(jtu.JaxTestCase):
 
   def setUp(self):
@@ -134,7 +134,6 @@ class JaxToTfTestCase(jtu.JaxTestCase):
     result_tf = None
 
     func_tf = jax2tf.convert(func_jax, enable_xla=enable_xla)
-    tf_args = _make_tf_args(args)
 
     unexpected_successes: List[str] = []
     # Run the "compiled" mode first, it is most important
@@ -150,7 +149,7 @@ class JaxToTfTestCase(jtu.JaxTestCase):
         continue
 
       try:
-        result_tf = _run_tf_function(func_tf, *tf_args, mode=mode)
+        result_tf = _run_tf_function(func_tf, *args, mode=mode)
         tf_exception = None
       except Exception as e:
         tf_exception = e
@@ -171,6 +170,10 @@ class JaxToTfTestCase(jtu.JaxTestCase):
             f"Unexpected success with known limitations {expect_tf_error}"))
           unexpected_successes.append(f"{mode}: {expect_tf_error}")
 
+      if (jtu.device_under_test() == "gpu" and
+          "dot_general_preferred" in self._testMethodName):
+        logging.info(log_message(f"Arguments are {args}, JAX result is {result_jax}\nand TF result is {result_tf}"))
+
       skip_comparison = [l for l in jax2tf_limits if l.skip_comparison]
       if skip_comparison:
         logging.warning(log_message(f"Skip result comparison due to {skip_comparison}"))
@@ -188,13 +191,65 @@ class JaxToTfTestCase(jtu.JaxTestCase):
       custom_assert_lim = [l for l in jax2tf_limits if l.custom_assert]
       assert len(custom_assert_lim) <= 1, f"Expecting at most one applicable limitation with custom_assert, found {custom_assert_lim}"
 
-      if custom_assert_lim:
-        logging.info(log_message(f"Running custom_assert with tol={max_tol} due to {custom_assert_lim[0]}"))
-        custom_assert_lim[0].custom_assert(self, result_jax, result_tf, args=args, tol=max_tol)
-      else:
-        logging.info(log_message(f"Running default assert with tol={max_tol}"))
-        # In compiled mode we expect the same result as JAX by default
-        self.assertAllClose(result_jax, result_tf, atol=max_tol, rtol=max_tol)
+      try:
+        err_msg = f"TF mode {mode}."
+        log_hlo_on_error = mode == "compiled" or jtu.device_under_test() == "tpu"
+        if log_hlo_on_error:
+          err_msg += " See the logs for JAX and TF HLO comparisons."
+        if custom_assert_lim:
+          logging.info(log_message(f"Running custom_assert with tol={max_tol} due to {custom_assert_lim[0]}"))
+          custom_assert_lim[0].custom_assert(self, result_jax, result_tf,
+                                             args=args, tol=max_tol,
+                                             err_msg=err_msg)
+        else:
+          logging.info(log_message(f"Running default assert with tol={max_tol}"))
+          self.assertAllClose(result_jax, result_tf, atol=max_tol, rtol=max_tol,
+                              err_msg=err_msg)
+      except AssertionError as e:
+        # Print the HLO for comparison
+        if not log_hlo_on_error:
+          print(f"[{self._testMethodName}] Not logging HLO because the "
+                f"mode was {mode}")
+          raise
+
+        logging.info(f"[{self._testMethodName}] Logging HLO for exception in mode {mode}: {e}")
+        jax_comp = jax.xla_computation(func_jax)(*args)
+        jax_hlo = jax_comp.as_hlo_text()
+        logging.info(f"[{self._testMethodName}] "
+                     f"JAX NON_OPT HLO\n{jax_hlo}")
+
+        tf_args_signature = _make_tf_input_signature(*args)
+        # If we give the signature, we cannot pass scalars
+        tf_args_no_scalars = tuple(
+            map(lambda a, sig: tf.convert_to_tensor(a, dtype=sig.dtype),
+                args, tf_args_signature))
+
+        tf_func_compiled = tf.function(
+            func_tf,
+            autograph=False,
+            jit_compile=True,
+            input_signature=tf_args_signature)
+        tf_hlo = tf_func_compiled.experimental_get_compiler_ir(*tf_args_no_scalars)(
+                    stage="hlo")
+        logging.info(f"[{self._testMethodName}] TF NON OPT HLO\n{tf_hlo}")
+
+        backend = jax.lib.xla_bridge.get_backend()
+        modules = backend.compile(jax_comp).hlo_modules()
+        jax_opt_hlo = modules[0].to_string()
+        logging.info(f"[{self._testMethodName}] "
+                     f"JAX OPT HLO\n{jax_opt_hlo}")
+
+        # TODO(b/189265364): Remove this workaround
+        if (jtu.device_under_test() == "gpu" and
+            "dot_general" in self._testMethodName):
+          print(f"[{self._testMethodName}] Not logging TF OPT HLO because of "
+                f"crash in tf.experimental_get_compiler_ir (b/189265364)")
+        else:
+          tf_opt_hlo = tf_func_compiled.experimental_get_compiler_ir(*tf_args_no_scalars)(
+                      stage="optimized_hlo")
+          logging.info(f"[{self._testMethodName}] TF OPT HLO\n{tf_opt_hlo}")
+
+        raise
 
     # end "for mode"
 
@@ -238,22 +293,26 @@ class JaxToTfTestCase(jtu.JaxTestCase):
       return self.ConvertAndCompare(grad_func, t_arg)
     assert False, transform
 
+
   def CheckShapePolymorphism(self, f_jax: Callable, *,
                              input_signature: Sequence[tf.TensorSpec],
                              polymorphic_shapes: Optional[Sequence[Any]],
-                             expected_output_signature: tf.TensorSpec):
-    """Convert a function using polymorphic shapes.
+                             expected_output_signature: Optional[tf.TensorSpec] = None,
+                             enable_xla: bool = True):
+    """Converts a function using polymorphic shapes.
 
     Args:
       f_jax: a JAX function of `n` arguments
       input_signature: used as the input signature for the tf.function.
-      in_shapes: if given, it must be a sequence of `n` shape specifications and
-        must match the `input_signature`. (see jax2tf.convert).
+      polymorphic_shapes: Specifies input shapes to be treated polymorphically
+        during conversion.
+      expected_output_signature: if given, this function tests whether the
+        actual output signature is equal to this one.
+      enable_xla: Whether to enable XLA conversion for jax2tf.convert.
     """
-    f_tf = tf.function(
-        jax2tf.convert(f_jax, polymorphic_shapes=polymorphic_shapes),
-        autograph=False,
-        input_signature=input_signature)
+    f_tf = jax2tf.convert(f_jax, polymorphic_shapes=polymorphic_shapes,
+                             enable_xla=enable_xla)
+    f_tf = tf.function(f_tf, autograph=False, input_signature=input_signature)
     concrete_f_tf = f_tf.get_concrete_function(*input_signature)
     if expected_output_signature:
       # Strangely, output_shapes can be a single shape for a function with a
@@ -284,3 +343,60 @@ class JaxToTfTestCase(jtu.JaxTestCase):
           dtype=tf.float32)
 
     return tree_util.tree_multimap(polymorphic_shape_to_tensorspec, polymorphic_shapes)
+
+
+  def CheckOpMetadata(self, jax_fun, x,
+                      expected: Sequence[OpMetadataGraph],
+                      include_xla_op_metadata=True):
+    """Checks that the tf.Graph obtained by converting `jax_fun` for argument
+    `x` contains all the given OpMetadata.
+
+    If `not include_xla_op_metadata` then disable the generation of the
+    OpMetadata attributes, and check that we don't find any ops with
+    metadata.
+    """
+    f_tf = tf.function(
+        jax2tf.convert(jax_fun,
+                       include_xla_op_metadata=include_xla_op_metadata),
+        autograph=False,
+        input_signature=[tf.TensorSpec(x.shape, x.dtype)])
+    # Trace the TF function to a graph
+    f_tf_concrete = f_tf.get_concrete_function(tf.convert_to_tensor(x))
+
+    found_tf_ops = []
+    def iter_nested_graph(graph: tf.Graph):
+      for n in graph._nodes_by_id.values():
+        try:
+          op_metadata = n.get_attr("_XlaOpMetadata")
+          op_metadata_proto = xla_data_pb2.OpMetadata()
+          op_metadata_proto.ParseFromString(op_metadata)
+          found_tf_ops.append(
+              OpMetadataGraph(
+                  tf_type=n.type,
+                  op_name=op_metadata_proto.op_name,
+                  op_type=op_metadata_proto.op_type,
+                  source_file=op_metadata_proto.source_file,
+                  source_line=op_metadata_proto.source_line))
+        except ValueError:
+          continue
+
+        # Look for nested graphs. There probably is a better way!
+        if n.type == "StatelessWhile":
+          iter_nested_graph(n._body_graph)
+          iter_nested_graph(n._cond_graph)
+        if n.type == "StatelessCase":
+          for idx in range(10):  # How can I tell how many cases there are?
+            branch = getattr(n, f"_branch_graph_{idx}", None)
+            if branch is None:
+              break
+            iter_nested_graph(branch)
+
+    iter_nested_graph(f_tf_concrete.graph)
+    try:
+      if include_xla_op_metadata:
+        self.assertContainsSubset(expected, found_tf_ops)
+      else:
+        self.assertEmpty(found_tf_ops)
+    except Exception:
+      print("Found nodes:\n  ", "\n   ".join([str(md) for md in found_tf_ops]))
+      raise
