@@ -12,6 +12,7 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+from enum import IntEnum
 import numpy as np
 from collections import OrderedDict, Counter
 from typing import Callable, Sequence, Tuple, Union
@@ -32,13 +33,14 @@ from ..errors import JAXTypeError
 from ..interpreters import ad
 from ..interpreters import pxla
 from ..interpreters import xla
+from ..interpreters import batching
 from ..interpreters import partial_eval as pe
 from ..lib import xla_bridge as xb
 from ..lib import xla_client as xc
 from ..tree_util import tree_flatten, tree_unflatten
 from .._src.util import (extend_name_stack, HashableFunction, safe_zip,
                          wrap_name, wraps, distributed_debug_log,
-                         split_list, cache)
+                         split_list, cache, tuple_insert)
 xops = xc._xla.ops
 
 def pjit(fun: Callable,
@@ -46,6 +48,115 @@ def pjit(fun: Callable,
          out_axis_resources,
          static_argnums: Union[int, Sequence[int]] = (),
          donate_argnums: Union[int, Sequence[int]] = ()):
+  """Makes ``fun`` compiled and automatically partitioned across multiple devices.
+
+  The returned function has semantics equivalent to those of ``fun``, but is
+  compiled to an XLA computation that runs across multiple devices
+  (e.g. multiple GPUs or multiple TPU cores). This can be useful if the jitted
+  version of ``fun`` would not fit in a single device's memory, or to speed up
+  ``fun`` by running each operation in parallel across multiple devices.
+
+  The partitioning over devices happens automatically based on
+  propagation of input partitioning specified in ``in_axis_resources`` and
+  output partitioning specified in ``out_axis_resources``. The resources
+  specified in those two arguments must refer to mesh axes, as defined by
+  the :py:func:`jax.experimental.maps.mesh` context manager. Note that the mesh
+  definition at ``pjit`` application time is ignored, and the returned function
+  will use the mesh definition available at each call site.
+
+  Inputs to a pjit'd function will be automatically partitioned across devices
+  if they're not already correctly partitioned based on ``in_axis_resources``.
+  In some scenarios, ensuring that the inputs are already correctly pre-partitioned
+  can increase performance. For example, if passing the output of one pjit'd function
+  to another pjit’d function (or the same pjit’d function in a loop), make sure the
+  relevant ``out_axis_resources`` match the corresponding ``in_axis_resources``.
+
+  .. note::
+    **Multi-process platforms:** On multi-process platforms such as TPU pods,
+    ``pjit`` can be used to run computations across all available devices across
+    processes. To achieve this, ``pjit`` is designed to be used in SPMD Python
+    programs, where every process is running the same Python code such that all
+    processes run the same pjit'd function in the same order.
+
+    When running in this configuration, the mesh should contain devices across
+    all processes. However, any input argument dimensions partitioned over
+    multi-process mesh axes should be of size equal to the corresponding *local*
+    mesh axis size, and outputs will be similarly sized according to the local
+    mesh. ``fun`` will still be executed across *all* devices in the mesh,
+    including those from other processes, and will be given a global view of the
+    data spread accross multiple processes as a single array. However, outside
+    of ``pjit`` every process only "sees" its local piece of the input and output,
+    corresponding to its local sub-mesh.
+
+    The SPMD model requires that the same multi-process ``pjit``'d functions must
+    be run in the same order on all processes, but they can be interspersed with
+    arbitrary operations running in a single process.
+
+  Args:
+    fun: Function to be compiled. Should be a pure function, as side-effects may
+      only be executed once. Its arguments and return value should be arrays,
+      scalars, or (nested) standard Python containers (tuple/list/dict) thereof.
+      Positional arguments indicated by ``static_argnums`` can be anything at
+      all, provided they are hashable and have an equality operation defined.
+      Static arguments are included as part of a compilation cache key, which is
+      why hash and equality operators must be defined.
+    in_axis_resources: Pytree of structure matching that of arguments to ``fun``,
+      with all actual arguments replaced by resource assignment specifications.
+      It is also valid to specify a pytree prefix (e.g. one value in place of a
+      whole subtree), in which case the leaves get broadcast to all values in
+      that subtree.
+
+      The valid resource assignment specifications are:
+        - :py:obj:`None`, in which case the value will be replicated on all devices
+        - :py:class:`PartitionSpec`, a tuple of length at most equal to the rank
+          of the partitioned value. Each element can be a :py:obj:`None`, a mesh
+          axis or a tuple of mesh axes, and specifies the set of resources assigned
+          to partition the value's dimension matching its position in the spec.
+
+      The size of every dimension has to be a multiple of the total number of
+      resources assigned to it.
+    out_axis_resources: Like ``in_axis_resources``, but specifies resource
+      assignment for function outputs.
+    static_argnums: An optional int or collection of ints that specify which
+      positional arguments to treat as static (compile-time constant).
+      Operations that only depend on static arguments will be constant-folded in
+      Python (during tracing), and so the corresponding argument values can be
+      any Python object.
+
+      Static arguments should be hashable, meaning both ``__hash__`` and
+      ``__eq__`` are implemented, and immutable. Calling the jitted function
+      with different values for these constants will trigger recompilation.
+      Arguments that are not arrays or containers thereof must be marked as
+      static.
+
+      If ``static_argnums`` is not provided, no arguments are treated as static.
+    donate_argnums: Specify which arguments are "donated" to the computation.
+      It is safe to donate arguments if you no longer need them once the
+      computation has finished. In some cases XLA can make use of donated
+      buffers to reduce the amount of memory needed to perform a computation,
+      for example recycling one of your input buffers to store a result. You
+      should not reuse buffers that you donate to a computation, JAX will raise
+      an error if you try to.
+
+  Returns:
+    A wrapped version of ``fun``, set up for just-in-time compilation and
+    automatic partitioned by the mesh available at each call site.
+
+  For example, a convolution operator can be automatically partitioned over
+  an arbitrary set of devices by a single ```pjit`` application:
+
+  >>> import jax
+  >>> import jax.numpy as jnp
+  >>> from jax.experimental.maps import mesh
+  >>> from jax.experimental.pjit import PartitionSpec, pjit
+  >>>
+  >>> x = jnp.arange(8, dtype=jnp.float32)
+  >>> f = pjit(lambda x: jax.numpy.convolve(x, jnp.asarray([0.5, 1.0, 0.5]), 'same'),
+  ...         in_axis_resources=None, out_axis_resources=PartitionSpec('devices'))
+  >>> with mesh(jax.devices(), ('devices',)):
+  ...   print(f(x))  # doctest: +SKIP
+  [ 0.5  2.   4.   6.   8.  10.  12.  10. ]
+  """
   warn("pjit is an experimental feature and probably has bugs!")
   _check_callable(fun)
 
@@ -81,6 +192,9 @@ def pjit(fun: Callable,
     # Putting this outside of wrapped would make resources lexically scoped
     resource_env = maps.thread_resources.env
     mesh = resource_env.physical_mesh
+    if mesh.empty:
+      raise RuntimeError("pjit requires a non-empty mesh! Are you sure that "
+                         "it's defined at the call site?")
 
     f = lu.wrap_init(fun)
     if static_argnums:
@@ -100,8 +214,8 @@ def pjit(fun: Callable,
     local_in_avals = tuple(core.raise_to_shaped(core.get_aval(a)) for a in args_flat)
     jaxpr, in_axis_resources_flat, out_axis_resources_flat = \
         _pjit_jaxpr(flat_fun, mesh, local_in_avals,
-                    in_tree, in_axis_resources,
-                    HashableFunction(out_tree, closure=()), out_axis_resources)
+                    in_tree, hashable_pytree(in_axis_resources),
+                    HashableFunction(out_tree, closure=()), hashable_pytree(out_axis_resources))
 
     out = pjit_p.bind(
         *args_flat,
@@ -118,12 +232,18 @@ def pjit(fun: Callable,
 class _ListWithW(list):
   __slots__ = ('__weakref__',)
 
+def hashable_pytree(pytree):
+  vals, treedef = tree_flatten(pytree)
+  vals = tuple(vals)
+  return HashableFunction(lambda: tree_unflatten(treedef, vals),
+                          closure=(treedef, vals))
+
 @lu.cache
 def _pjit_jaxpr(fun, mesh, local_in_avals,
-                in_tree, in_axis_resources,
-                out_tree, out_axis_resources):
+                in_tree, in_axis_resources_thunk,
+                out_tree, out_axis_resources_thunk):
   in_axis_resources_flat = tuple(flatten_axes("pjit in_axis_resources",
-                                              in_tree, in_axis_resources))
+                                              in_tree, in_axis_resources_thunk()))
   _check_shapes_against_resources("pjit arguments", False, mesh.local_mesh.shape,
                                   local_in_avals, in_axis_resources_flat)
   global_in_avals = local_to_global(mesh, local_in_avals, in_axis_resources_flat)
@@ -132,17 +252,47 @@ def _pjit_jaxpr(fun, mesh, local_in_avals,
   jaxpr = core.ClosedJaxpr(jaxpr, consts)
 
   out_axis_resources_flat = tuple(flatten_axes("pjit out_axis_resources",
-                                               out_tree(), out_axis_resources))
+                                               out_tree(), out_axis_resources_thunk()))
   _check_shapes_against_resources("pjit outputs", mesh.is_multi_process, mesh.shape,
                                   global_out_avals, out_axis_resources_flat)
   # lu.cache needs to be able to create weakrefs to outputs, so we can't return a plain tuple
   return _ListWithW([jaxpr, in_axis_resources_flat, out_axis_resources_flat])
 
 
+class SpecSync(IntEnum):
+  """Encodes how much out of sync the real value of partitions is compared to the user specified one.
+
+  We use this to make sure we don't show garbage modified values while claiming
+  that the users have specified them like that.
+  """
+  DIM_PERMUTE = 1  # Dimensions permuted, but no new sharding axes
+  IN_SYNC = 2  # Entirely in sync
+
 class ParsedPartitionSpec:
-  def __init__(self, user_spec, partitions):
+  __slots__ = ('partitions', 'unsafe_user_spec', 'sync')
+
+  def __init__(self, user_spec, partitions, sync=SpecSync.IN_SYNC):
     self.partitions = tuple(partitions)
-    self.user_spec = user_spec
+    self.unsafe_user_spec = user_spec
+    self.sync = sync
+
+  @property
+  def user_spec(self):
+    return self.unsynced_user_spec(SpecSync.IN_SYNC)
+
+  def unsynced_user_spec(self, min_sync):
+    if self.sync < min_sync:
+      raise AssertionError(f"Please open a bug report! ({self.sync} >= {min_sync})")
+    return self.unsafe_user_spec
+
+  def insert_axis_partitions(self, dim, val):
+    parts = self.partitions
+    too_short = dim - len(parts)
+    if too_short > 0:
+      parts += ((),) * too_short
+    new_partitions = tuple_insert(parts, dim, val)
+    new_sync = SpecSync.DIM_PERMUTE if val == () else SpecSync.IN_SYNC
+    return ParsedPartitionSpec(self.unsafe_user_spec, new_partitions, sync=new_sync)
 
   @classmethod
   def from_user_input(cls, entry, arg_name):
@@ -163,13 +313,12 @@ class ParsedPartitionSpec:
     return cls(entry, axis_specs)
 
   def __hash__(self):
-    return hash(self.partitions)
+    return hash((self.partitions, self.sync))
 
   def __eq__(self, other):
-    return (self.partitions, self.user_spec) == (other.partitions, other.user_spec)
-
-  def __str__(self):
-    return str(self.user_spec)
+    return (self.partitions == other.partitions and
+            self.unsafe_user_spec == other.unsafe_user_spec and
+            self.sync == other.sync)
 
   def __len__(self):
     return len(self.partitions)
@@ -196,11 +345,12 @@ def _check_unique_resources(axis_resources, arg_name):
   for arg_axis_resources in axis_resources:
     if not arg_axis_resources: continue
     resource_counts = Counter(it.chain.from_iterable(arg_axis_resources))
+    if not resource_counts: continue
     if resource_counts.most_common(1)[0][1] > 1:
       multiple_uses = [r for r, c in resource_counts.items() if c > 1]
       if multiple_uses:
         raise ValueError(f"A single {arg_name} specification can map every mesh axis "
-                         f"to at most one positional dimension, but {arg_axis_resources} "
+                         f"to at most one positional dimension, but {arg_axis_resources.user_spec} "
                          f"has duplicate entries for {maps.show_axes(multiple_uses)}")
 
 def _check_shapes_against_resources(what: str, is_global_shape: bool, mesh_shape, flat_avals, flat_axis_resources):
@@ -209,7 +359,7 @@ def _check_shapes_against_resources(what: str, is_global_shape: bool, mesh_shape
     shape = aval.shape
     if len(shape) < len(aval_axis_resources):
       raise ValueError(f"One of {what} was given the resource assignment "
-                       f"of {aval_axis_resources!s}, which implies that "
+                       f"of {aval_axis_resources.user_spec}, which implies that "
                        f"it has a rank of at least {len(aval_axis_resources)}, "
                        f"but it is {len(shape)}")
     for i, axis_resources in enumerate(aval_axis_resources):
@@ -217,11 +367,11 @@ def _check_shapes_against_resources(what: str, is_global_shape: bool, mesh_shape
         size = int(np.prod([mesh_shape[resource] for resource in axis_resources], dtype=np.int64))
       except KeyError as e:
         raise ValueError(f"One of {what} was given the resource assignment "
-                         f"of {aval_axis_resources!s}, but resource axis "
+                         f"of {aval_axis_resources.user_spec}, but resource axis "
                          f"{e.args[0]} is undefined. Did you forget to declare the mesh?") from None
       if shape[i] % size != 0:
         raise ValueError(f"One of {what} was given the resource assignment "
-                         f"of {aval_axis_resources!s}, which implies that "
+                         f"of {aval_axis_resources.user_spec}, which implies that "
                          f"the{global_str} size of its dimension {i} should be "
                          f"divisible by {size}, but it is equal to {shape[i]}")
 
@@ -299,6 +449,41 @@ def _pjit_translation_rule(c, axis_env, in_nodes, name_stack, backend, name,
 xla.call_translations[pjit_p] = _pjit_translation_rule
 
 
+def _pjit_batcher(insert_axis,
+                  vals_in, dims_in,
+                  axis_name, main_type,
+                  jaxpr, in_axis_resources, out_axis_resources,
+                  resource_env, donated_invars, name):
+  axis_size, = {x.shape[d] for x, d in zip(vals_in, dims_in) if d is not batching.not_mapped}
+  # batch_jaxpr expects all batching dimensions to be equal to 0
+  vals_in = [batching.moveaxis(x, d, 0) if d is not batching.not_mapped and d != 0
+             else x for x, d in zip(vals_in, dims_in)]
+  is_mapped_in = [d is not batching.not_mapped for d in dims_in]
+  new_jaxpr, is_mapped_out = batching.batch_jaxpr(
+      jaxpr, axis_size, is_mapped_in,
+      instantiate=False, axis_name=axis_name, main_type=main_type)
+
+  new_parts = (axis_name,) if insert_axis else ()
+  in_axis_resources = tuple(
+      spec.insert_axis_partitions(0, new_parts) if is_mapped else spec
+      for is_mapped, spec in zip(is_mapped_in, in_axis_resources))
+  out_axis_resources = tuple(
+      spec.insert_axis_partitions(0, new_parts) if is_mapped else spec
+      for is_mapped, spec in zip(is_mapped_out, out_axis_resources))
+  vals_out = pjit_p.bind(
+    *vals_in,
+    jaxpr=new_jaxpr,
+    in_axis_resources=in_axis_resources,
+    out_axis_resources=out_axis_resources,
+    resource_env=resource_env,
+    donated_invars=donated_invars,
+    name=name)
+  dims_out = [0 if batched else batching.not_mapped for batched in is_mapped_out]
+  return vals_out, dims_out
+batching.initial_style_batchers[pjit_p] = partial(_pjit_batcher, False)
+pxla.spmd_primitive_batchers[pjit_p] = partial(_pjit_batcher, True)
+
+
 def _pjit_jvp(primals_in, tangents_in,
               jaxpr, in_axis_resources, out_axis_resources,
               resource_env, donated_invars, name):
@@ -334,7 +519,8 @@ def _check_resources_against_named_axes(what, aval, pos_axis_resources, named_ax
   overlap = pjit_resources & aval_resources
   if overlap:
     raise JAXTypeError(
-        f"{what} has an axis resources specification of {pos_axis_resources} "
+        f"{what} has an axis resources specification of "
+        f"{pos_axis_resources.unsynced_user_spec(SpecSync.DIM_PERMUTE)} "
         f"that uses one or more mesh axes already used by xmap to partition "
         f"a named axis appearing in its named_shape (both use mesh axes "
         f"{maps.show_axes(overlap)})")
@@ -407,13 +593,20 @@ def get_array_mapping(axis_resources: ParsedPartitionSpec) -> pxla.ArrayMapping:
                      for i, axes in enumerate(axis_resources)
                      for axis in axes)
 
-def get_sharding_proto(c, xla_op, axis_resources, mesh):
+def get_sharding_proto(c, xla_op, axis_resources: ParsedPartitionSpec,
+                       mesh: maps.Mesh) -> xc.OpSharding:
   xla_shape = c.GetShape(xla_op)
   if xla_shape.is_token():
     aval = core.abstract_token
     assert axis_resources is REPLICATED
   else:
-    aval = core.ShapedArray(xla_shape.dimensions(), xla_shape.element_type())
+    aval = core.ShapedArray(xla_shape.dimensions(), xla_shape.element_type())  # type: ignore
+  return get_aval_sharding_proto(aval, axis_resources, mesh)
+
+
+def get_aval_sharding_proto(aval: core.AbstractValue,
+                            axis_resources: ParsedPartitionSpec,
+                            mesh: maps.Mesh) -> xc.OpSharding:
   array_mapping = get_array_mapping(axis_resources)
   sharding_spec = pxla.mesh_sharding_specs(mesh.shape, mesh.axis_names)(
       aval, array_mapping)
