@@ -58,6 +58,10 @@ def _check_jaxlib_version():
 
 _check_jaxlib_version()
 
+if version >= (0, 1, 68):
+  from jaxlib import cpu_feature_guard
+  cpu_feature_guard.check_cpu_features()
+
 from jaxlib import xla_client
 from jaxlib import lapack
 from jaxlib import pocketfft
@@ -99,9 +103,9 @@ except ImportError:
 _xla_extension_version = getattr(xla_client, '_version', 0)
 
 try:
-  from jaxlib import tpu_client  # pytype: disable=import-error
+  from jaxlib import tpu_client as tpu_driver_client  # pytype: disable=import-error
 except:
-  tpu_client = None
+  tpu_driver_client = None
 
 
 cuda_path: Optional[str]
