@@ -31,6 +31,8 @@ logging._warn_preinit_stderr = 0
 
 import jax.lib
 from .._src.config import flags
+from . import tpu_driver_client
+from . import xla_client
 from jax._src import util, traceback_util
 from jax._src import dtypes
 import numpy as np
@@ -38,29 +40,25 @@ import threading
 
 traceback_util.register_exclusion(__file__)
 
-try:
-  from . import tpu_client
-except ImportError:
-  tpu_client = None
-from . import xla_client
 
 xops = xla_client.ops
 
 FLAGS = flags.FLAGS
 
+# TODO(phawkins): Remove jax_xla_backend.
 flags.DEFINE_string(
-    'jax_xla_backend', 'xla',
-    'Default is "xla" for the XLA service directly, '
-    'or "tpu_driver" for using high-performance access to Cloud TPU hardware.')
+    'jax_xla_backend', '',
+    'jax_xla_backend is an alias for jax_platform_name. If both are '
+    'provided, --jax_xla_backend takes priority. Prefer --jax_platform_name.')
 flags.DEFINE_string(
     'jax_backend_target', 'local',
     'Either "local" or "rpc:address" to connect to a remote service target.')
 flags.DEFINE_string(
     'jax_platform_name',
     os.getenv('JAX_PLATFORM_NAME', ''),
-    'Platform name for XLA. The default is to attempt to use a GPU if '
+    'Platform name for XLA. The default is to attempt to use a GPU or TPU if '
     'available, but fall back to CPU otherwise. To set the platform manually, '
-    'pass "cpu" for CPU or "gpu" for GPU.')
+    'pass "cpu" for CPU, "gpu" for GPU, etc.')
 flags.DEFINE_bool(
     'jax_disable_most_optimizations', False,
     'Try not to do much optimization work. This can be useful if the cost of '
@@ -126,51 +124,100 @@ def get_compile_options(
 
   return compile_options
 
-_backends = {}
 
-def register_backend(name, factory):
-  _backends[name] = factory
+# Backends
 
-def _get_local_backend(platform=None):
-  if not platform:
-    platform = FLAGS.jax_platform_name or None
-
-  backend = xla_client.get_local_backend(platform)
-  if backend is None:
-    raise RuntimeError("No local XLA backends found.")
-
-  if backend.platform == 'cpu' and platform != 'cpu':
-    logging.warning('No GPU/TPU found, falling back to CPU. '
-                    '(Set TF_CPP_MIN_LOG_LEVEL=0 and rerun for more info.)')
-
-  return backend
+def _make_tpu_driver_client():
+  if tpu_driver_client is None:
+    logging.info("Remote TPU is not linked into jax; skipping remote TPU.")
+    return None
+  if FLAGS.jax_backend_target is None:
+    logging.info("No --jax_backend_target was provided; skipping remote TPU.")
+    return None
+  return tpu_driver_client.TpuBackend.create(worker=FLAGS.jax_backend_target)
 
 
-register_backend('xla', _get_local_backend)
+# Backends, in increasing order of preference.
+# We have no particular opinion about how "backends" relate to "devices". For
+# example, there could be multiple backends that provide the same kind of
+# device.
+_backend_factories = {}
 
-# memoize the TPU driver to be consistent with xla_client behavior
-_tpu_backend = None
-
-def _get_tpu_driver_backend(platform):
-  if platform == "cpu":
-    return _get_local_backend("cpu")
-
-  global _tpu_backend
-  if _tpu_backend is None:
-    backend_target = FLAGS.jax_backend_target
-    if backend_target is None:
-      raise ValueError('When using TPU Driver as the backend, you must specify '
-                       '--jax_backend_target=<hostname>:8470.')
-    _tpu_backend = tpu_client.TpuBackend.create(worker=backend_target)
-  return _tpu_backend
+def register_backend_factory(name, factory, *, priority=0):
+  _backend_factories[name] = (factory, priority)
 
 
-if tpu_client:
-  register_backend('tpu_driver', _get_tpu_driver_backend)
+if jax.lib._xla_extension_version >= 23:
+   register_backend_factory('interpreter', xla_client.make_interpreter_client,
+                            priority=-100)
+   register_backend_factory('cpu', xla_client.make_cpu_client,
+                            priority=0)
+   register_backend_factory('tpu_driver', _make_tpu_driver_client,
+                            priority=100)
+   register_backend_factory('gpu', xla_client.make_gpu_client,
+                            priority=200)
+   register_backend_factory('tpu', xla_client.make_tpu_client,
+                            priority=300)
+else:
+   register_backend_factory('interpreter',
+                            xla_client._interpreter_backend_factory,
+                            priority=-100)
+   register_backend_factory('cpu', xla_client._cpu_backend_factory, priority=0)
+   register_backend_factory('tpu_driver', _make_tpu_driver_client,
+                            priority=100)
+   register_backend_factory('gpu', xla_client._gpu_backend_factory,
+                            priority=200)
+   register_backend_factory('tpu', xla_client._tpu_backend_factory,
+                            priority=300)
 
 
+_default_backend = None
+_backends = None
 _backend_lock = threading.Lock()
 
+
+def backends():
+  global _backends
+  global _default_backend
+
+  with _backend_lock:
+    if _backends is not None:
+      return _backends
+
+    default_priority = -1000
+    _backends = {}
+    for name, (factory, priority) in _backend_factories.items():
+      logging.vlog(1, "Initializing backend '%s'" % name)
+      try:
+        backend = factory()
+        if backend is not None:
+          if backend.device_count() > 0:
+            _backends[name] = backend
+          util.distributed_debug_log(("Initialized backend", backend.platform),
+                                     ("process_index", backend.process_index()),
+                                     ("device_count", backend.device_count()),
+                                     ("local_devices", backend.local_devices()))
+          logging.vlog(1, "Backend '%s' initialized" % name)
+          if priority > default_priority:
+            _default_backend = backend
+            default_priority = priority
+      except Exception as err:
+        if name in ('cpu', 'interpreter'):
+          # We always expect the CPU and interpreter backends to initialize
+          # successfully.
+          raise
+        else:
+          # If the backend isn't built into the binary, or if it has no devices,
+          # we expect a RuntimeError.
+          logging.info("Unable to initialize backend '%s': %s" % (name, err))
+          continue
+    if _default_backend.platform == "cpu":
+      logging.warning('No GPU/TPU found, falling back to CPU. '
+                      '(Set TF_CPP_MIN_LOG_LEVEL=0 and rerun for more info.)')
+    return _backends
+
+
+
 @lru_cache(maxsize=None)  # don't use util.memoize because there is no X64 dependence.
 def get_backend(platform=None):
   # TODO(mattjj,skyewm): remove this input polymorphism after we clean up how
@@ -178,23 +225,28 @@ def get_backend(platform=None):
   if not isinstance(platform, (type(None), str)):
     return platform
 
-  with _backend_lock:
-    backend_factory = _backends.get(FLAGS.jax_xla_backend)
-    if backend_factory is None:
-      msg = 'Unknown jax_xla_backend value "{}".'
-      raise ValueError(msg.format(FLAGS.jax_xla_backend))
-    backend = backend_factory(platform)
-    util.distributed_debug_log(("Initialized backend", backend.platform),
-                               ("process_index", backend.process_index()),
-                               ("device_count", backend.device_count()),
-                               ("local_devices", backend.local_devices()))
+  bs = backends()
+  platform = (platform or FLAGS.jax_xla_backend or FLAGS.jax_platform_name
+              or None)
+  if platform is not None:
+    backend = bs.get(platform, None)
+    if backend is None:
+      raise RuntimeError(f"Unknown backend {platform}")
     return backend
+  else:
+    return _default_backend
 
 
 def get_device_backend(device=None):
   """Returns the Backend associated with `device`, or the default Backend."""
-  platform = device.platform if device else None
-  return get_backend(platform)
+  if device is not None:
+    # TODO(phawkins): remove this workaround after jaxlib 0.1.68 becomes the
+    # minimum and it is safe to call `.client` on a tpu_driver TpuDevice.
+    if tpu_driver_client and isinstance(
+        device, tpu_driver_client._tpu_client.TpuDevice):
+      return get_backend('tpu_driver')
+    return device.client
+  return get_backend()
 
 
 def device_count(backend: Optional[str] = None) -> int:
@@ -354,7 +406,7 @@ def normalize_to_xla_dtypes(val):
 def _numpy_array_constant(builder, value, canonicalize_types=True):
   if canonicalize_types:
     value = normalize_to_xla_dtypes(value)
-  return xops.ConstantLiteral(builder, value)
+  return [xops.ConstantLiteral(builder, value)]
 
 def parameter(builder, num, shape, name=None, replicated=None):
   if name is None:
@@ -369,6 +421,22 @@ def parameter(builder, num, shape, name=None, replicated=None):
                         replicated)
 
 
+def constant_general(builder, py_val, canonicalize_types=True):
+  """Translate a general constant `py_val` to a constant, canonicalizing its dtype.
+
+  Args:
+    py_val: a Python value to be translated to a constant.
+
+  Returns:
+    A representation of the constant as a list of xla ops.
+  """
+  for t in type(py_val).mro():
+    handler = _constant_handlers.get(t)
+    if handler: return handler(builder, py_val, canonicalize_types)
+  if hasattr(py_val, '__jax_array__'):
+    return constant(builder, py_val.__jax_array__(), canonicalize_types)
+  raise TypeError("No constant handler for type: {}".format(type(py_val)))
+
 def constant(builder, py_val, canonicalize_types=True):
   """Translate constant `py_val` to a constant, canonicalizing its dtype.
 
@@ -378,12 +446,9 @@ def constant(builder, py_val, canonicalize_types=True):
   Returns:
     A representation of the constant, either a ComputationDataHandle or None
   """
-  for t in type(py_val).mro():
-    handler = _constant_handlers.get(t)
-    if handler: return handler(builder, py_val, canonicalize_types)
-  if hasattr(py_val, '__jax_array__'):
-    return constant(builder, py_val.__jax_array__(), canonicalize_types)
-  raise TypeError("No constant handler for type: {}".format(type(py_val)))
+  const = constant_general(builder, py_val, canonicalize_types=canonicalize_types)
+  assert len(const) == 1, f"Internal error: cannot create constant from object of type {type(py_val)}"
+  return const[0]
 
 # HLO instructions optionally can be annotated to say how the output should be
 # spatially partitioned (represented in XLA as OpSharding protos, see
@@ -484,10 +549,10 @@ def _ndarray_constant_handler(c, val, canonicalize_types=True):
     collapsed_val = val[tuple(0 if ax in zero_stride_axes else slice(None)
                               for ax in range(val.ndim))]
     xla_val = xops.Broadcast(
-        _numpy_array_constant(c, collapsed_val, canonicalize_types),
+        _numpy_array_constant(c, collapsed_val, canonicalize_types)[0],
         np.take(val.shape, zero_stride_axes))
     permutation = np.argsort(tuple(zero_stride_axes) + tuple(other_axes))
-    return xops.Transpose(xla_val, permutation)
+    return [xops.Transpose(xla_val, permutation)]
   else:
     return _numpy_array_constant(c, val, canonicalize_types)
 register_constant_handler(np.ndarray, _ndarray_constant_handler)
