@@ -14,5 +14,6 @@
 
 # flake8: noqa: F401
 from jax._src.ops.scatter import (
-  index, index_add, index_mul, index_update, index_min, index_max, segment_sum
+  index, index_add, index_mul, index_update, index_min, index_max,
+  segment_sum, segment_prod, segment_min, segment_max,
 )
