@@ -14,4 +14,4 @@
 
 # flake8: noqa: F401
 
-from jax._src.scipy.optimize.minimize import minimize
+from jax._src.scipy.optimize.minimize import (minimize, OptimizeResults)
