@@ -17,7 +17,7 @@ import functools
 import re
 import os
 import textwrap
-from typing import Dict, Sequence, Union
+from typing import Dict, List, Generator, Sequence, Tuple, Union
 import unittest
 import warnings
 import zlib
@@ -33,10 +33,12 @@ from . import core
 from ._src import dtypes as _dtypes
 from . import lax
 from ._src.config import flags, bool_env, config
-from ._src.util import partial, prod
+from ._src.util import partial, prod, unzip2
 from .tree_util import tree_multimap, tree_all, tree_map, tree_reduce
 from .lib import xla_bridge
 from .interpreters import xla
+from .experimental import maps
+from .experimental.maps import mesh
 
 
 FLAGS = flags.FLAGS
@@ -138,7 +140,10 @@ def _assert_numpy_allclose(a, b, atol=None, rtol=None, err_msg=''):
   kw = {}
   if atol: kw["atol"] = atol
   if rtol: kw["rtol"] = rtol
-  np.testing.assert_allclose(a, b, **kw, err_msg=err_msg)
+  with np.errstate(invalid='ignore'):
+    # TODO(phawkins): surprisingly, assert_allclose sometimes reports invalid
+    # value errors. It should not do that.
+    np.testing.assert_allclose(a, b, **kw, err_msg=err_msg)
 
 def tolerance(dtype, tol=None):
   tol = {} if tol is None else tol
@@ -781,7 +786,11 @@ def assert_dot_precision(expected_precision, fun, *args):
                 if eqn.primitive == lax.dot_general_p]
   for precision in precisions:
     msg = "Unexpected precision: {} != {}".format(expected_precision, precision)
-    assert precision == expected_precision, msg
+    if isinstance(precision, tuple):
+      assert precision[0] == expected_precision, msg
+      assert precision[1] == expected_precision, msg
+    else:
+      assert precision == expected_precision, msg
 
 
 _CACHED_INDICES: Dict[int, Sequence[int]] = {}
@@ -862,20 +871,22 @@ class JaxTestCase(parameterized.TestCase):
   def rng(self):
     return self._rng
 
-  def assertArraysEqual(self, x, y, *, check_dtypes=True):
+  def assertArraysEqual(self, x, y, *, check_dtypes=True, err_msg=''):
     """Assert that x and y arrays are exactly equal."""
     if check_dtypes:
       self.assertDtypesMatch(x, y)
-    np.testing.assert_array_equal(x, y)
+    # Work around https://github.com/numpy/numpy/issues/18992
+    with np.errstate(over='ignore'):
+      np.testing.assert_array_equal(x, y, err_msg=err_msg)
 
   def assertArraysAllClose(self, x, y, *, check_dtypes=True, atol=None,
-                           rtol=None):
+                           rtol=None, err_msg=''):
     """Assert that x and y are close (up to numerical tolerances)."""
     self.assertEqual(x.shape, y.shape)
     atol = max(tolerance(_dtype(x), atol), tolerance(_dtype(y), atol))
     rtol = max(tolerance(_dtype(x), rtol), tolerance(_dtype(y), rtol))
 
-    _assert_numpy_allclose(x, y, atol=atol, rtol=rtol)
+    _assert_numpy_allclose(x, y, atol=atol, rtol=rtol, err_msg=err_msg)
 
     if check_dtypes:
       self.assertDtypesMatch(x, y)
@@ -888,27 +899,30 @@ class JaxTestCase(parameterized.TestCase):
       self.assertEqual(_dtype(x), _dtype(y))
 
   def assertAllClose(self, x, y, *, check_dtypes=True, atol=None, rtol=None,
-                     canonicalize_dtypes=True):
+                     canonicalize_dtypes=True, err_msg=''):
     """Assert that x and y, either arrays or nested tuples/lists, are close."""
     if isinstance(x, dict):
       self.assertIsInstance(y, dict)
       self.assertEqual(set(x.keys()), set(y.keys()))
       for k in x.keys():
         self.assertAllClose(x[k], y[k], check_dtypes=check_dtypes, atol=atol,
-                            rtol=rtol, canonicalize_dtypes=canonicalize_dtypes)
+                            rtol=rtol, canonicalize_dtypes=canonicalize_dtypes,
+                            err_msg=err_msg)
     elif is_sequence(x) and not hasattr(x, '__array__'):
       self.assertTrue(is_sequence(y) and not hasattr(y, '__array__'))
       self.assertEqual(len(x), len(y))
       for x_elt, y_elt in zip(x, y):
         self.assertAllClose(x_elt, y_elt, check_dtypes=check_dtypes, atol=atol,
-                            rtol=rtol, canonicalize_dtypes=canonicalize_dtypes)
+                            rtol=rtol, canonicalize_dtypes=canonicalize_dtypes,
+                            err_msg=err_msg)
     elif hasattr(x, '__array__') or np.isscalar(x):
       self.assertTrue(hasattr(y, '__array__') or np.isscalar(y))
       if check_dtypes:
         self.assertDtypesMatch(x, y, canonicalize_dtypes=canonicalize_dtypes)
       x = np.asarray(x)
       y = np.asarray(y)
-      self.assertArraysAllClose(x, y, check_dtypes=False, atol=atol, rtol=rtol)
+      self.assertArraysAllClose(x, y, check_dtypes=False, atol=atol, rtol=rtol,
+                                err_msg=err_msg)
     elif x == y:
       return
     else:
@@ -1000,6 +1014,44 @@ def ignore_warning(**kw):
     warnings.filterwarnings("ignore", **kw)
     yield
 
+# -------------------- Mesh parametrization helpers --------------------
+
+MeshSpec = List[Tuple[str, int]]
+
+@contextmanager
+def with_mesh(named_shape: MeshSpec) -> Generator[None, None, None]:
+  """Test utility for setting up meshes given mesh data from `schedules`."""
+  # This is similar to the `with_mesh` function above, but isn't a decorator.
+  axis_names, shape = unzip2(named_shape)
+  size = prod(shape)
+  local_devices = list(api.local_devices())
+  if len(local_devices) < size:
+    raise unittest.SkipTest(f"Test requires {size} local devices")
+  mesh_devices = np.array(local_devices[:size]).reshape(shape)
+  with mesh(mesh_devices, axis_names):
+    yield
+
+def with_mesh_from_kwargs(f):
+  return lambda *args, **kwargs: with_mesh(kwargs['mesh'])(f)(*args, **kwargs)
+
+def with_and_without_mesh(f):
+  return parameterized.named_parameters(
+    {"testcase_name": name, "mesh": mesh, "axis_resources": axis_resources}
+    for name, mesh, axis_resources in (
+      ('', (), ()),
+      ('Mesh', (('x', 2),), (('i', 'x'),))
+    ))(with_mesh_from_kwargs(f))
+
+old_spmd_lowering_flag = False
+def set_spmd_lowering_flag(val: bool):
+  global old_spmd_lowering_flag
+  maps.make_xmap_callable.cache_clear()
+  old_spmd_lowering_flag = maps.EXPERIMENTAL_SPMD_LOWERING
+  maps.EXPERIMENTAL_SPMD_LOWERING = val
+
+def restore_spmd_lowering_flag():
+  maps.make_xmap_callable.cache_clear()
+  maps.EXPERIMENTAL_SPMD_LOWERING = old_spmd_lowering_flag
 
 class _cached_property:
   null = object()
